import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from .errors import NonNumericDataError, SampleFullOfNaNs
import statsmodels.api as sm


class PMM(BaseEstimator, TransformerMixin):
    def __init__(self):
        """
        Initializing the algorithm.
        """
        return

    def fit(self, X: 'np.array', y: 'np.array' = None, **fit_params: dict):
        """
        :param X: np.array of shape (n_samples, n_features)
            The input samples (optional).
        :param y: np.array of shape (n_samples,), default = None
            The target values (optional).
        :param fit_params: dict, default = None
            Additional fit parameters (optional).
        :return: self
        """
        return self

    def __get_chi_squared_value(self, X: 'np.array'):
        '''
        Function to calculate the chi-value of an array.
        :param X: np.array of shape (n_samples, n_features)
            The input samples (required).
        :return: the chi-value of the x_observed array.
        '''
        X_chi = np.multiply.outer(X.sum(1), X.sum(0)) / X.sum()
        X_chi = ((X_chi - X) ** 2) / X_chi
        return X_chi.sum()

    def transform(self, X: 'np.array', y: 'np.array' = None, **fit_params: dict):
        """
        Filling in the missing values of X.
        :param X: np.array of shape (n_samples, n_features)
            The input samples.
        :param y: np.array of shape (n_samples, ), default = None
            The target values (none for unsupervised transformations).
        :param fit_params: dict
            Optional additional fit parameters.
        :return: np.array of shape (n_samples, n_features)
            The original data, but with the missing values imputed.
        """

        # checking if received data is ok
        if not np.issubdtype(X.dtype, np.number):
            raise NonNumericDataError("The given DataFrame contains non-numeric values.")

        # making the working copy of X and its missing values
        X_copy = X.copy()
        missing_values = np.where(pd.isnull(X_copy))

        if len(set(missing_values[1])) >= X_copy.shape[1] - 1:
            raise SampleFullOfNaNs("All columns contain missing values. Can't impute using pmm. ")

        # a buffer where the columns with missing values are kept
        columns_still_with_missing_values = list(set(missing_values[1]))

        while len(columns_still_with_missing_values) > 0:
            x = np.delete(X_copy, columns_still_with_missing_values, axis = 1)
            column = columns_still_with_missing_values.pop()
            y = X_copy[:, column]

            missing_values_rows = missing_values[0][np.where(missing_values[1] == column)]
            x_missing = x[missing_values_rows]
            x_observed = np.delete(x, missing_values_rows, axis=0)
            y_observed = np.delete(y, missing_values_rows, axis=0)

            ols = sm.OLS(y_observed, sm.add_constant(x_observed)).fit()
            beta_shapka = np.delete(ols.params, 0, axis = 0)
            # sigma_shapka = ols.params[0]
            epsilon_shapka = abs(ols.predict(sm.add_constant((x_observed))) - y_observed)
            std_dev = np.matmul(epsilon_shapka.T, epsilon_shapka) / self.__get_chi_squared_value(x_observed)
            beta_zviozdacika = np.random.multivariate_normal(
                beta_shapka,
                std_dev * np.linalg.inv(np.matmul(x_observed.T, x_observed))
            )

            y_observed_pred = np.matmul(x_observed, beta_shapka)
            y_missing_pred = np.matmul(x_missing, beta_zviozdacika)
            for i in range(len(y_missing_pred)):
                deltas = np.vstack(
                    (abs(y_observed_pred - y_missing_pred[i]), y_observed_pred)
                ).T
                deltas = deltas[deltas[:, 0].argsort()]
                y[missing_values_rows[i]] = np.random.choice(deltas[:3, 1])

        return X_copy

    def fit_transform(self, X: 'np.array', y: 'np.array' = None, **fit_params: dict):
        """
        Fits transformer to X and y with optional parameters fit_params and returns a transformed version of X.
        :param X: np.array of shape (n_samples, n_features)
            The input samples.
        :param y: np.array of shape (n_samples, ), default = None
            The target values (none for unsupervised transformations).
        :param fit_params: dict
            Optional additional fit parameters.
        :return: np.array of shape (n_samples, n_features)
            The original data, but with the missing values imputed.
        """
        return self.fit(X).transform(X)

    def apply(self, df: 'pd.DataFrame', columns: list):
        """
        Apply the algorithm on a DataFrame.
        :param df: pandas DataFrame
            The DataFrame with possible NaN-values that should be imputed
        :param columns: list of str
            The columns on which to apply the algorithm.
        :return: pandas DataFrame
            The DataFrame with imputed missing-values.
        """
        if not set(columns).issubset(set(df.columns)):
            raise ValueError("Columns not found in the DataFrame.")

        data = df[columns].copy()
        imputed_data = self.fit_transform(data.to_numpy())

        df.update(pd.DataFrame(imputed_data, columns=columns))



'''
Created with love by Sigmoid
@Author - Basoc Nicoleta-Nina - nicoleta.basoc28@gmail.com
'''

import pandas as pd
import numpy as np
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import train_test_split
from sklearn.base import BaseEstimator, TransformerMixin

from .errors import OnlyOneColumnSelectedError, NonNumericDataError


class MICE(BaseEstimator, TransformerMixin):
    
    def fit(self, X: 'np.array', y: 'np.array'= None, **fit_params):
        """
        :param X: array-like of shape (n_samples, n_features)
            input samples
        :param y: array-like of shape (n_samples)
            target values (None for unsupervised transformations)
        :param fit_params: dict
            additional fit parameters
        :return: self
        """
        return self

    def transform(self, X: 'np.array', y: 'np.array' = None, **fit_params):

        if not np.issubdtype(X.dtype, np.number):
            raise NonNumericDataError('The given array contains non-numeric values !')

        df_copy = pd.DataFrame(X).copy()
        new_df = []

        for column in df_copy.columns:
            if df_copy[column].isna().any():
                # Select columns all the other columns except the one we want to impute right now
                columns_imputed = df_copy.loc[:, df_copy.columns != column].columns
                # Select the indexes of NaN elements
                nan_indexes = np.where(df_copy[column].isna())[0]
                # Perform mean imputation and create the zeroth dataset
                zeroth_df = pd.DataFrame(np.array([df_copy[column_zeroth].fillna(np.mean(df_copy[column_zeroth]))
                                                   for column_zeroth in columns_imputed]).transpose())
                # Select train and test indicies
                train_indicies = df_copy.loc[pd.isna(df_copy[column]) == False].index
                test_indicies = df_copy.loc[pd.isna(df_copy[column])].index
                # Train linear regression
                linear_regression = LinearRegression()
                linear_regression.fit(zeroth_df.iloc[train_indicies], df_copy[column].iloc[train_indicies])

                predictions = linear_regression.predict(zeroth_df.iloc[test_indicies])
                # Impute the predictions to the given positions
                for idx, prediction in zip(test_indicies, predictions):
                    df_copy[column].iloc[idx] = prediction
            else:
                continue

        return df_copy.values
    
    def fit_transform(self, X: 'np.array', y: 'np.array'=None, **fit_params):
        """
        Fits transformer to X and y with optional parameters fit_params and returns a transformed version of X.
        :param X: array-like of shape (n_samples, n_features)
            input samples
        :param y: array-like of shape (n_samples,)
            target values (None for unsupervised transformations)
        :param fit_params: dict
            additional fit parameters
        :return: ndarray array of shape (n_samples, n_features_new)
            transformed array
        """

        return self.fit(X).transform(X)

    def apply(self, df: 'pandas DataFrame', columns: list):
        """
        Apply the algorithm on a DataFrame
        :param df: pandas DataFrame
            the DataFrame with the possible NaN-values that should be imputed
        :param columns: list of strings
            the column that should be taken into account during the NaN-values imputation process
        :return:
        """
        #Only one column selected
        if len(df[columns].columns)<=1:
            raise OnlyOneColumnSelectedError('The given algorithm cannot be applied only on one column')
        
        # check if the given columns are in the Dataframe
        cols_not_in_df = [col for col in columns if col not in df]
        if cols_not_in_df:
            raise ValueError(f"{', '.join(cols_not_in_df)} are not present in the DataFrame")

        # select the desired columns
        sub_df = df[columns]
        
        # check if data is numeric
        for col in sub_df.columns:
            if not pd.api.types.is_numeric_dtype(sub_df[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # convert DataFrame to numpy array
        sub_df = sub_df.to_numpy()

        # apply the transformation on the DataFrame
        sub_df_new = self.fit_transform(sub_df)

        # save the transformation in the DataFrame to return
        df.update(pd.DataFrame(sub_df_new, columns=columns))



"""
Created with love by Sigmoid
@Author - Sclearuc Marius - marius.sclearuc@gmail.com
"""

import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from .MICE import MICE


class SICE(BaseEstimator, TransformerMixin):
    def __init__(self, n_cycle : int = 5):
        """
            Initializing the algorithm.
        :param n_cycle: int, default = 0.05
            The number of cycles the algorithm goes through.
        """
        self.n_cycle = n_cycle

    def fit(self, X: 'np.array', y: 'np.array' = None, **fit_params: dict):
        """
        :param X: np.array of shape (n_samples, n_features)
            The input samples (optional).
        :param y: np.array of shape (n_samples,), default = None
            The target values (optional).
        :param fit_params: dict, default = None
            Additional fit parameters (optional).
        :return: self
        """
        return self

    def transform(self, X: 'np.array', y: 'np.array' = None, **fit_params: dict):
        """
        Fill in the missing values of X.
        :param X: np.array of shape (n_samples, n_features)
            The input samples.
        :param y: np.array of shape (n_samples, ), default = None
            The target values (none for unsupervised transformations).
        :param fit_params: dict
            Optional additional fit parameters.
        :return: np.array of shape (n_samples, n_features)
            The original data, but with the missing values imputed.
        """
        missing_values = np.where(pd.isnull(X))
        missing_values = list(zip(missing_values[0], missing_values[1]))
        mice_computed_values = {location: list() for location in missing_values}

        mice = MICE()

        for i in range(self.n_cycle):
            mice_output = mice.fit_transform(X)
            for location in missing_values:
                mice_computed_values[location].append(mice_output[missing_values])

        X_copy = X.copy()
        for location in missing_values:
            if type(mice_computed_values[location][0]) in [int, float]:
                X_copy[location] = np.mean(mice_computed_values[location])
            else:
                values, counts = np.unique(mice_computed_values[location], return_counts=True)
                X_copy[location] = values[np.argmax(counts)]

        return X_copy

    def fit_transform(self, X, y=None, **fit_params):
        """
        Fits transformer to X and y with optional parameters fit_params and returns a transformed version of X.
        :param X: np.array of shape (n_samples, n_features)
            The input samples.
        :param y: np.array of shape (n_samples, ), default = None
            The target values (none for unsupervised transformations).
        :param fit_params: dict
            Optional additional fit parameters.
        :return: np.array of shape (n_samples, n_features)
            The original data, but with the missing values imputed.
        """
        return self.fit(X).transform(X)

    def apply(self, df, columns):
        """
        Apply the algorithm on a DataFrame.
        :param df: pandas DataFrame
            The DataFrame with possible NaN-values that should be imputed
        :param columns: list of str
            The columns on which to apply the algorithm.
        :return: pandas DataFrame
            The DataFrame with imputed missing-values.
        """
        data = df[columns].copy()
        imputed_data = self.transform(data.values)

        df.update(pd.DataFrame(imputed_data, columns=columns))



'''
Created with love by Sigmoid
@Author - Denis Smocvin - denissmocvin@gmail.com
'''

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.neighbors import NearestNeighbors

import pandas as pd
import numpy as np

from .errors import NonNumericDataError, NoSuchDistanceMetricError, AllSamplesHaveNaNsError, SampleFullOfNaNs


class CDI(BaseEstimator, TransformerMixin):
    def __init__(self, n_neighbors: int = 5, metric: str = 'minkowski', p: int = 2, n_select: int = 1):
        '''
        Initialize the algorithm.
        :param n_neighbors: the number of neighbors to find for every sample with missing values
        :param metric: the distance metric to use for finding the nearest neighbor
        :param p: the parameter for the Minkowski metric
        :param n_select: the index of the selected sample
        '''

        self.n_neighbors = n_neighbors
        self.metric = metric
        self.p = p
        self.n_select = n_select

        self.DISTANCES = ['euclidean', 'manhattan', 'chebyshev', 'minkowski', 'wminkowski', 'seuclidean', 'mahalanobis']

        if metric.lower() not in self.DISTANCES:
            error_text = f"{metric} metric isn\'t supported right now, choose one of the following: {', '.join(self.DISTANCES)}"
            raise NoSuchDistanceMetricError(error_text)

        if  n_select > n_neighbors:
            error_text = f'n_select ({n_select}) > n_neighbors ({n_neighbors}) : The selected sample index is higher than the n neighbors used by the algorithm, but should be lower.'
            raise ValueError(error_text)

    def fit(self, X: 'np.array', y: 'np.array'=None, **fit_params):
        """
        :param X: array-like of shape (n_samples, n_features)
            input samples
        :param y: array-like of shape (n_samples,)
            target values (None for unsupervised transformations)
        :param fit_params: dict
            additional fit parameters
        :return: self
        """

        return self

    def transform(self, X: 'np.array', y: 'np.array'=None, **fit_params):
        """
        Impute according to the Cold Deck Imputation algorithm. It will try to find as many neighbors as possible up to
        the given n_neighbors number of neighbors and choose the n_select-th one for imputation.

        Note: there should be at least on row without NaN values.
        :param X: array-like of shape (n_samples, n_features)
            input samples
        :param y: array-like of shape (n_samples,)
            target values (None for unsupervised transformations)
        :param fit_params: dict
            additional fit parameters
        :return: ndarray array of shape (n_samples, n_features_new)
            transformed array
        """

        # check if data is numeric
        if not np.issubdtype(X.dtype, np.number):
            raise NonNumericDataError('The given array contains non-numeric values !')

        X_new = X.copy()

        for i, sample in enumerate(X_new):
            if np.isnan(sample).all():
                error_text = f'Sample {i} is full of NaNs. Cannot operate with such samples. Consider removing the row.'
                raise SampleFullOfNaNs(error_text)

            if np.isnan(sample).any():
                # create a copy of the original array with deleted rows that contained nans
                X_temp = X_new.copy()
                X_temp = np.array([x for x in X_temp if not np.isnan(x).any()])

                if len(X_temp) == 0:
                    error_text = 'All samples contain NaN values. Cold Deck Imputation isn\'t fitted for this configuration. Consider using other imputation algorithms.'
                    raise AllSamplesHaveNaNsError(error_text)

                # the algorithm will try to find as many neighbors as possible, the limit being the given number of neighbors
                n_neighbors = self.n_neighbors
                if n_neighbors > len(X_temp):
                    n_neighbors = len(X_temp)

                # put at the start of the array the sample requiring imputation with means temporarily in place of nans
                temp = sample.copy()
                means = [col.mean() for col in X_temp.T]

                for j, _ in enumerate(temp):
                    if np.isnan(temp[j]):
                        temp[j] = means[j]

                X_temp = np.insert(X_temp, 0, temp, axis=0)

                # find the nearest neighbors
                knn = NearestNeighbors()
                knn.fit(X_temp)
                neighbors = knn.kneighbors(return_distance=False, n_neighbors=n_neighbors)

                # impute
                neighbor = X_temp[neighbors[0][self.n_select - 1]]

                for j, _ in enumerate(sample):
                    if np.isnan(sample[j]):
                        sample[j] = neighbor[j]

        return X_new

    def fit_transform(self, X: 'np.array', y: 'np.array'=None, **fit_params):
        """
        Fits transformer to X and y with optional parameters fit_params and returns a transformed version of X.
        :param X: array-like of shape (n_samples, n_features)
            input samples
        :param y: array-like of shape (n_samples,)
            target values (None for unsupervised transformations)
        :param fit_params: dict
            additional fit parameters
        :return: ndarray array of shape (n_samples, n_features_new)
            transformed array
        """

        return self.fit(X).transform(X)

    def apply(self, df: 'pandas DataFrame', columns: list):
        """
        Apply the algorithm on a DataFrame
        :param df: pandas DataFrame
            the DataFrame with the possible NaN-values that should be imputed
        :param columns: list of strings
            the column that should be taken into account during the NaN-values imputation process
        :return:
        """

        # check if the given columns are in the DataFrame
        cols_not_in_df = [col for col in columns if col not in df]
        if cols_not_in_df:
            raise ValueError(f"{', '.join(cols_not_in_df)} are not present in the DataFrame")

        # select the desired columns
        sub_df = df[columns]

        # check if data is numeric
        for col in sub_df.columns:
            if not pd.api.types.is_numeric_dtype(sub_df[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # convert DataFrame to numpy array
        sub_df = sub_df.to_numpy()

        # apply the transformation
        sub_df_new = self.fit_transform(sub_df)

        # save the transformation in the DataFrame
        df.update(pd.DataFrame(sub_df_new, columns=columns))


"""
Created with love by Sigmoid
@Author - Sclearuc Marius - marius.sclearuc@gmail.com
"""

import pandas as pd
import numpy as np
from random import choice
from sklearn.base import BaseEstimator, TransformerMixin

from .errors import NoSuchDistanceMetricError, NonNumericDataError


class HotDeckImputation(BaseEstimator, TransformerMixin):
    def __init__(self, n_neighbors: int = 5, metric: str = 'minkowski', p: int = 2, eps: float = 1e-10):
        """
            Initializing the algorithm.
        :param n_neighbors: int, default = 0.05
            The number of neighbours to consider for KNN.
        :param metric: str, default = 'minkowski'
            The type of distance metric to use for calculating the distance between data.
            Note: Supported metrics are: chebyshev, euclidean, seuclidean, manhattan, mahalanobis, minkowski.
        :param p: int, default = 2
            The parameter for the Minkowski metric.
            Note: for p = 2, the 'minkowski' metric is equivalent to the 'euclidean' metric.
        :param eps: float, default = 1e-10
            The additional error.
        """
        self.n_neighbors = n_neighbors
        self.p = p
        self.eps = eps

        # checking if metric distance is supported
        if metric in ['euclidean', 'manhattan', 'chebyshev', 'minkowski', 'wminkowski', 'seuclidean', 'mahalanobis']:
            self.metric = metric
        else:
            raise NoSuchDistanceMetricError(
                "X metric isn't supported right now, choose on of the following: euclidean, manhattan, "
                "chebyshev, minkowski, wminkowski, seuclidean, mahalanobis.")

    def fit(self, X: 'np.array', y: 'np.array' = None, **fit_params: dict):
        """
        :param X: np.array of shape (n_samples, n_features)
            The input samples (optional).
        :param y: np.array of shape (n_samples,), default = None
            The target values (optional).
        :param fit_params: dict, default = None
            Additional fit parameters (optional).
        :return: self
        """
        return self

    def __get_distance(self, a: 'np.array', b: 'np.array'):
        """
        Function to calculate the distance according to the metric.
        :param a: np.array
            The first array.
        :param b: np.array
            The second array.
        :return: float
            The computed distance.
        """
        if a.shape != b.shape:
            raise ValueError("Mismatch in __get_distance.")
        elif self.metric == 'manhattan':
            return sum([abs(a[i] - b[i]) for i in range(len(a))])
        elif self.metric == 'euclidean':
            return sum([(a[i] - b[i]) ** 2 for i in range(len(a))]) ** 0.5
        elif self.metric == 'chebyshev':
            return max([(a[i] - b[i]) for i in range(len(a))])
        elif self.metric == 'minkowski':
            return sum([abs(a[i] - b[i]) ** self.p for i in range(len(a))]) ** (1/self.p)
        elif self.metric == 'wminkowski':
            return -1
        elif self.metric == 'seuclidean':
            return sum([(a[i] - b[i]) ** 2 for i in range(len(a))])
        cov_matrix = np.cov(np.stack((a, b), axis=1))
        if np.linalg.det(cov_matrix) == 0:
            raise ValueError("Can't compute covariance matrix.")
        inv_cov_matrix = np.linalg.inv(cov_matrix)
        return np.matmul(np.matmul((a - b).T, inv_cov_matrix), (a - b))

    def transform(self, X, y=None, **fit_params):
        """
        Fill in the missing values of X.
        :param X: np.array of shape (n_samples, n_features)
            The input samples.
        :param y: np.array of shape (n_samples, ), default = None
            The target values (none for unsupervised transformations).
        :param fit_params: dict
            Optional additional fit parameters.
        :return: np.array of shape (n_samples, n_features)
            The original data, but with the missing values imputed.
        """
        # checking if received data is ok
        if not np.issubdtype(X.dtype, np.number):
            raise NonNumericDataError("The given DataFrame contains non-numeric values.")

        # checking if there are enough rows to find at least self.n_neighbors values
        if X.shape[0] < self.n_neighbors + 1:
            raise ValueError("The number of specified neighbors exceeds the shape number of samples.")

        # creating the working copy and making a dict mapping id (i.e. rows) to the columns of the missing values
        # i.e. if for row nr.2 there are missing values in columns 4, 10, then dict[2] = [4, 10]
        X_copy = X.copy()
        missing_values = np.where(np.isnan(X))
        missing_values = list(zip(missing_values[0], missing_values[1]))
        id_to_missing_value_columns_map = {i: list() for i in range(X.shape[0])}
        for id, column in missing_values:
            id_to_missing_value_columns_map[id].append(int(column))

        for id, column in missing_values:
            # taking a record/row and eliminating all its missing columns
            current_record = np.delete(X[id], id_to_missing_value_columns_map[id])
            # making a copy of all the dataset less the current record
            records = np.delete(X, id, axis=0)
            distances = list()
            for record in range(records.shape[0]):
                # checking if it's possible to compare the distances of the 2 records
                if np.isnan(records[record][column]) or \
                        (not set(id_to_missing_value_columns_map[record]).issubset(set(id_to_missing_value_columns_map[id]))):
                    continue
                # computing the distance, then appending it to the list of all distances
                distance = self.__get_distance(
                    current_record,
                    np.delete(
                        records[record],
                        id_to_missing_value_columns_map[id]
                    )
                )
                distances.append((records[record][column], distance))

            # checking once again if there are enough records to impute a missing value
            if len(distances) < 1:
                raise ValueError("Too many lacking spaces, can't properly fill the data.")
            distances.sort(key=lambda elem: elem[1])

            # choosing a random near neighbour and imputing its value
            nearest = distances[:min(len(distances), self.n_neighbors)]
            randomly_chosen_value = choice(nearest)
            X_copy[id, column] = randomly_chosen_value[0] + self.eps

        return X_copy

    def fit_transform(self, X: 'np.array', y: 'np.array' = None, **fit_params):
        """
        Fits transformer to X and y with optional parameters fit_params and returns a transformed version of X.
        :param X: np.array of shape (n_samples, n_features)
            The input samples.
        :param y: np.array of shape (n_samples, ), default = None
            The target values (none for unsupervised transformations).
        :param fit_params: dict
            Optional additional fit parameters.
        :return: np.array of shape (n_samples, n_features)
            The original data, but with the missing values imputed.
        """
        return self.fit(X).transform(X)

    def apply(self, df: 'pd.DataFrame', columns: list):
        """
        Apply the algorithm on a DataFrame.
        :param df: pandas DataFrame
            The DataFrame with possible NaN-values that should be imputed
        :param columns: list of str
            The columns on which to apply the algorithm.
        :return: pandas DataFrame
            The DataFrame with imputed missing-values.
        """
        if not set(columns).issubset(set(df.columns.values)):
            raise ValueError("Columns not found in the DataFrame.")

        imputed_data = self.fit_transform(df[columns].to_numpy())

        df.update(pd.DataFrame(imputed_data, columns=columns))



class NonNumericDataError(Exception):
    pass

class NoSuchDistanceMetricError(Exception):
    pass

class AllSamplesHaveNaNsError(Exception):
    pass

class SampleFullOfNaNs(Exception):
    pass

class OnlyOneColumnSelectedError(Exception):
    pass


'''
Created with love by Sigmoid
@Author - Denis Smocvin - denissmocvin@gmail.com
'''

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.neighbors import NearestNeighbors
import pandas as pd
import numpy as np

from .errors import NonNumericDataError, NoSuchDistanceMetricError, AllSamplesHaveNaNsError, SampleFullOfNaNs

from scipy.spatial.distance import minkowski

class FRNNI(BaseEstimator, TransformerMixin):
    def __init__(self, n_neighbors: int = 5, metric: str = 'minkowski', p: int =2):
        '''
        Initialize the algorithm.
        :param n_neighbors: the number of neighbors to find for every sample with missing values
        :param metric: the distance metric to use for finding the nearest neighbors
        :param p: the parameter for the Minkowski metric
        '''

        self.distances = ['euclidean', 'manhattan', 'chebyshev', 'minkowski', 'wminkowski', 'seuclidean', 'mahalanobis']
        self.n_neighbors = n_neighbors
        self.metric = metric
        self.p = p

        if metric.lower() not in self.distances:
            raise NoSuchDistanceMetricError(f'{metric} metric isn\'t supported right now, choose one of the following: euclidean, manhattan, chebyshev, minkowski, wminkowski, seuclidean, mahalanobis.')

    def fit(self, X: 'np.array', y: 'np.array'=None, **fit_params):
        """
        :param X: array-like of shape (n_samples, n_features)
            input samples
        :param y: array-like of shape (n_samples,)
            target values (None for unsupervised transformations)
        :param fit_params: dict
            additional fit parameters
        :return: self
        """

        return self

    def transform(self, X: 'np.array', y: 'np.array' = None, **fit_params):
        """
        fill the missing values using Fuzzy-Rough Nearest Neighbor algorithm for imputation
        :param X: array-like of shape (n_samples, n_features)
            2d matrix representing the data that should be filled
        :param y: array-like of shape (n_samples,)
            1d array representing the target vector
        :param fit_params: dict
            the fit parameteres of the function
        :return: ndarray array of shape (n_samples, n_features_new)
            transformed array
        """

        # check if data is numeric
        if not np.issubdtype(X.dtype, np.number):
            raise NonNumericDataError('The given array contains non-numeric values !')

        X_new = X.copy()

        for i, sample in enumerate(X_new):
            if np.isnan(sample).all():
                error_text = f'Sample {i} is full of NaNs. Cannot operate with such samples. Consider removing the row.'
                raise SampleFullOfNaNs(error_text)

            if np.isnan(sample).any():
                # find neighbors
                d, N = self.__getNearestNeighbors(sample, X, n_neighbors=self.n_neighbors, metric=self.metric)

                for j, a in enumerate(sample):
                    if np.isnan(a):
                        tau1 = tau2 = 0

                        # create a copy of the original array with deleted rows that contained nans
                        X_temp = X_new.copy()
                        X_temp = np.array([x for x in X_temp if not np.isnan(x).any()])

                        if len(X_temp) == 0:
                            error_text = 'All samples contain NaN values. FRNNI isn\'t fitted for this configuration.'
                            raise AllSamplesHaveNaNsError(error_text)

                        # print(N)
                        for k, z in enumerate(N):
                            for l, t in enumerate(N):
                                # print(z, t)
                                if np.array_equal(t, z):
                                    continue

                                dist = (d[k], d[l])
                                # print(d, dist)

                                lower = self.__approximation(t, z, np.array(X_new).T[j], dist, N, j, type='lower')
                                upper = self.__approximation(t, z, np.array(X_new).T[j], dist, N, j, type='upper')

                                M = (lower + upper) / 2

                                tau1 += M * z[j]
                                tau2 += M

                        sample[j] = tau1 / tau2 if tau2 > 0 else sum([neighbor[j] for neighbor in N]) / len(N)

        return X_new

    def fit_transform(self, X: 'np.array', y: 'np.array'=None, **fit_params):
        """
        Fits transformer to X and y with optional parameters fit_params and returns a transformed version of X.
        :param X: array-like of shape (n_samples, n_features)
            2d matrix representing the data that should be filled
        :param y: array-like of shape (n_samples,)
            1d array representing the target vector
        :param fit_params: dict
            the fit parameteres of the function
        :return: ndarray array of shape (n_samples, n_features_new)
            transformed array
        """

        return self.fit(X).transform(X)

    def apply(self, df: 'pandas DataFrame', columns: 'list'):
        """
        Apply the algorithm on a DataFrame
        :param df: pandas DataFrame
            the DataFrame with the possible NaN-values that should be imputed
        :param columns: list of strings
            the column that should be taken into account during the NaN-values imputation process
        :return:
        """

        # check if the given columns are in the DataFrame
        cols_not_in_df = [col for col in columns if col not in df]
        if cols_not_in_df:
            raise ValueError(f"{', '.join(cols_not_in_df)} are not present in the DataFrame")

        # select the desired columns
        sub_df = df[columns]

        # check if data is numeric
        for col in sub_df.columns:
            if not pd.api.types.is_numeric_dtype(sub_df[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values !')

        # convert DataFrame to numpy array
        sub_df = sub_df.to_numpy()

        # apply the transformation
        sub_df_new = self.fit_transform(sub_df)

        # save the transformation in the DataFrame
        df.update(pd.DataFrame(sub_df_new, columns=columns))

    def __getNearestNeighbors(self, sample, X, n_neighbors, metric):
        # make a copy and drop nans
        X_temp = X.copy()
        X_temp = np.array([x for x in X_temp if not np.isnan(x).any()])

        # the algorithm will try to find as many neighbors as possible, the limit being the given number of neighbors
        if n_neighbors > len(X_temp):
            n_neighbors = len(X_temp)

        # put at the start of the array the sample requiring imputation with means temporarily in place of nans
        temp = sample.copy()
        means = [col.mean() for col in X_temp.T]

        for j, _ in enumerate(temp):
            if np.isnan(temp[j]):
                temp[j] = means[j]

        X_temp = np.insert(X_temp, 0, temp, axis=0)

        # find the nearest neighbors
        if metric == 'minkowski':
            knn = NearestNeighbors(metric=metric, p=self.p)
        else:
            knn = NearestNeighbors(metric=metric)
        knn.fit(X_temp)
        dist, neighbors = knn.kneighbors(n_neighbors=n_neighbors)

        return dist[0], X_temp[neighbors[0]]

    def __R_a(self, y, t, a):
        return 1 - np.abs(y - t) / (max(a) - min(a))

    def __R(self, y, t, a, d):
        equal = 1 if np.array_equal(d[0], d[1]) else 0

        return min(self.__R_a(y, t, a), 1 - equal)

    def __implication(self, a, b):
        return max(1 - a, b)

    def __t_norm(self, a, b):
        return min(a, b)

    def __approximation(self, y: list, z: list, a: list, d: tuple, N: 'np.array', i: int, type: str):
        values = []

        for t in N:
            r = self.__R(y[i], t[i], a, d)
            r_a = self.__R_a(y[i], z[i], a)

            if type == 'lower':
                values.append(self.__implication(r, r_a))
            elif type == 'upper':
                values.append(self.__t_norm(r, r_a))

        return min(values) if type == 'lower' else max(values)